"""Module for Creating a Client

Functions:
    thread
    create_client
"""

from __future__ import annotations

import asyncio
import threading
from typing import Callable
from urllib.parse import urlparse

from penne.delegates import Delegate
from penne.core import Client 

def thread(loop: asyncio.AbstractEventLoop, client: Client):
    """Method for starting background thread

    Args:
        loop (event loop): event loop used for new thread
        client (client object): client used for websocket connection
    """

    try:
        asyncio.set_event_loop(loop)
        loop.run_until_complete(client.run())
    except Exception as e:
        print(f"Connection terminated: {e}")


def create_client(address: str, custom_delegate_hash: dict[str, Delegate] = {}, on_connected: Callable=None):
    """Create a client object and start background thread

    Args:
        address (str): 
            url for connecting to server
        custom_delegate_hash (dict):
            mapping specifiers to new delegate methods to override default
        on_connected (callable):
            function to be called once connection is established

    Raises:
        ValueError: Address given must be a websocket
    """

    # Process address
    address_parts = urlparse(address)
    if address_parts.scheme not in ["ws", "wss"]:
        raise ValueError("Address given must be a websocket!")

    loop = asyncio.new_event_loop()
    

    # Create client instance and thread
    client = Client(address, loop, custom_delegate_hash, on_connected)
    t = threading.Thread(target=thread, args=(loop, client))
 
    client.thread = t
    client.thread.start()
 
    return client


"""Module Defining Message Related Classes"""

from __future__ import annotations
from typing import Any

class Message(object):
    """Generic Message Class for casting messages sent from server

    Attributes:
        dynamically added to each instance using the from dict method
    """

    @classmethod
    def from_dict(cls, raw_dict: dict[str, Any]):
        """Function to generate generic message objects
        
        Recursively catches nested structures

        Parameters:
            raw_dict (dict) : dict to be cast as message object
        """

        processed_dict = {}
        for key, value in raw_dict.items():

            # Recursive case
            if isinstance(value, dict):
                processed_dict[key] = cls.from_dict(value)

            # Recursive case for list of objects
            elif value and isinstance(value, list) and isinstance(value[0], dict):
                obj_list = []
                for nested_dict in value:
                    obj_list.append(cls.from_dict(nested_dict))
                processed_dict[key] = obj_list
            
            # Cast ID's to tuples - careful could be off - list of two values
            elif isinstance(value, list) and len(value) == 2:
                processed_dict[key] = tuple(value)

            else:
                processed_dict[key] = value

        obj = cls()
        obj.__dict__.update(processed_dict)
        return obj
    

    def as_dict(self):
        """Get dictionary representation of class"""

        rep = self.__dict__
        for key, value in rep.items():

            # Recursive case
            if isinstance(value, Message):
                rep[key] = value.as_dict()

        return rep


    def __getitem__(self, __name: str):
        return getattr(self, __name)


    def __repr__(self):
        return f"{self.__dict__}"


class HandleInfo(object):
    """Class to organize useful info for processing each type of message

    Attributes:
        specifier (str) : keyword for delegate and state maps
        action (str)    : action performed by message
    """

    def __init__(self, specifier, action):
        self.specifier = specifier
        self.action = action



"""Module with Core Implementation of Client"""

from __future__ import annotations

import asyncio
from typing import Any
import websockets
from cbor2 import loads, dumps

from . import messages, handlers, delegates


default_delegates = {
    "entities" : delegates.EntityDelegate,
    "tables" : delegates.TableDelegate,
    "plots" : delegates.PlotDelegate,
    "signals" : delegates.SignalDelegate,
    "methods" : delegates.MethodDelegate,
    "materials" : delegates.MaterialDelegate,
    "geometries" : delegates.GeometryDelegate,
    "lights" : delegates.LightDelegate,
    "images" : delegates.ImageDelegate,
    "textures" : delegates.TextureDelegate,
    "samplers" : delegates.SamplerDelegate,
    "buffers" : delegates.BufferDelegate,
    "bufferviews" : delegates.BufferViewDelegate,
    "document" : delegates.DocumentDelegate
}

class Client(object):
    """Client for communicating with server

    Attributes:
        _url (string): 
            address used to connect to server
        _loop (event loop): 
            event loop used for network thread
        delegates (dict): 
            map for delegate functions     
        is_connected(threading.Event): 
            signal when connection is ready
        thread (thread object): 
            network thread used by client
        _socket (WebSocketClientProtocol): 
            socket to connect to server
        name (str): 
            name of the client
        state (dict): 
            dict keeping track of created objects
        client_message_map (dict): 
            mapping message type to corresponding id
        server_message_map (dict):
            mapping message id's to handling info
        current_invoke (str):  
            id for next method invoke
        callback_map (dict): 
            mapping invoke_id to callback function
    """

    def __init__(self, url: str, loop, custom_delegate_hash: dict[str, delegates.Delegate], on_connected=None):
        """Constructor for the Client Class

        Args:
            _url (string): 
                address used to connect to server
            loop (event loop): 
                event loop used for network thread
            custom_delegate_hash (dict): 
                map for new delegate methods
            is_connected (threading Event):
                event object signaling when connection is ready
        """

        self._url = url
        self._loop = loop
        self.on_connected = on_connected
        self.delegates = {}
        self.thread = None
        self._socket = None
        self.name = "Python Client"
        self.state = {
            "entities": {},
            "tables": {},
            "plots": {},
            "signals": {},
            "methods": {},
            "materials": {},
            "geometries": {},
            "lights": {},
            "images": {},
            "textures": {},
            "samplers": {},
            "buffers": {},
            "bufferviews": {}
        }
        self.client_message_map = {
            "intro": 0,
            "invoke": 1
        }
        self.server_message_map = {
            0 : messages.HandleInfo("methods", "create"),
            1 : messages.HandleInfo("methods", "delete"),
            2 : messages.HandleInfo("signals", "create"),
            3 : messages.HandleInfo("signals", "delete"),
            4 : messages.HandleInfo("entities", "create"),
            5 : messages.HandleInfo("entities", "update"),
            6 : messages.HandleInfo("entities", "delete"),
            7 : messages.HandleInfo("plots", "create"),
            8 : messages.HandleInfo("plots", "update"),
            9 : messages.HandleInfo("plots", "delete"),
            10 : messages.HandleInfo("buffers", "create"),
            11 : messages.HandleInfo("buffers", "delete"),
            12 : messages.HandleInfo("bufferviews", "create"),
            13 : messages.HandleInfo("bufferviews", "delete"),
            14 : messages.HandleInfo("materials", "create"),
            15 : messages.HandleInfo("materials", "update"),
            16 : messages.HandleInfo("materials", "delete"),
            17 : messages.HandleInfo("images", "create"),
            18 : messages.HandleInfo("images", "delete"),
            19 : messages.HandleInfo("textures", "create"), 
            20 : messages.HandleInfo("textures", "delete"),
            21 : messages.HandleInfo("samplers", "create"),
            22 : messages.HandleInfo("samplers", "delete"),
            23 : messages.HandleInfo("lights", "create"),
            24 : messages.HandleInfo("lights", "update"),
            25 : messages.HandleInfo("lights", "delete"),
            26 : messages.HandleInfo("geometries", "create"),
            27 : messages.HandleInfo("geometries", "delete"),
            28 : messages.HandleInfo("tables", "create"),
            29 : messages.HandleInfo("tables", "update"),
            30 : messages.HandleInfo("tables", "delete"),
            31 : messages.HandleInfo("document", "update"),
            32 : messages.HandleInfo("document", "reset"),  
            33 : messages.HandleInfo("signals", "invoke"),  
            34 : messages.HandleInfo("methods", "reply"),
            35 : messages.HandleInfo("document", "initialized")
        }
        self._current_invoke = 0
        self.callback_map = {}

        # Hook up delegate map to default or custom based on input hash
        for key in default_delegates:
            if key not in custom_delegate_hash:
                self.delegates[key] = default_delegates[key]
            else:
                self.delegates[key] = custom_delegate_hash[key]
        self.state["document"] = self.delegates["document"](self, None, "document")


    def object_from_name(self, name: str, specifier: str) -> list[int]:
        """Get a method's id from its name

        Args:
            name (str): name of method

        Returns:
            Id group attached to the method

        Raises:
            Couldn't find method exception
        """
        objects: list[delegates.MethodDelegate] = self.state[specifier].values()
        for object in objects:
            if hasattr(object.info, "name") and object.info.name == name:
                return object.info.id
        raise Exception(f"Couldn't find object '{name}' in {specifier}")
            

    def invoke_method(self, id, args: list, context: dict[str, tuple] = None, on_done = None):
        """Invoke method on server

        Constructs a dictionary of arguments to use in send_message. The
        Dictionary follows the structure of an InvokeMethodMessage, but
        using a dictionary prevents from converting back and forth just
        before sending.

        Also implements callback functions attached to each invocation. By
        default, each invocation will store a None object in the callback
        map, and the handler responsible for reply messages will delete pop
        it from the map and call the method if there is one

        Args:
            id (list or str): 
                id or name for method
            args (list): 
                arguments for method
            context (dict): 
                optional, target context for method call
            callback (function): 
                function to be called upon response
        """
        
        # Get proper ID
        if isinstance(id, str):
            id = self.object_from_name(id, "methods")

        # Get invoke ID
        invoke_id = str(self._current_invoke)
        self._current_invoke += 1

        # Keep track of callback
        self.callback_map[invoke_id] = on_done

        # Construct message dict
        arg_dict = {
            "method": id,
            "args": args,
            "invoke_id": invoke_id
        }
        if context: arg_dict["context"] = context
        
        self.send_message(arg_dict, "invoke")


    def send_message(self, message_dict: dict[str, Any], type: str):
        """Send message to server

        Args:
            message_dict (dict): dict mapping message attribute to value
            type (str): either 'invoke' or 'intro' to indicate type of client message
        """

        # Construct message with ID from map and converted message object
        message = [self.client_message_map[type], message_dict]
        #print(f"Sending Message: {message}")
        
        asyncio.run_coroutine_threadsafe(self._socket.send(dumps(message)), self._loop)


    async def run(self):
        """Network thread for managing websocket connection"""  

        async with websockets.connect(self._url) as websocket:

            # update class
            self._socket = websocket
            self.name = (f"Python Client @ {self._url}")

            # send intro message
            intro = {"client_name": self.name}
            self.send_message(intro, "intro")

            # decode, iterate over, and handle all incoming messages
            async for message in self._socket:
                raw_message = loads(message)
                iterator = iter(raw_message)
                for id in iterator:
                    try:
                        handlers.handle(self, id, next(iterator))
                    except Exception as e:
                        print(f"Exception: {e}")

    
    def show_methods(self):
        """Displays Available Methods to the User"""

        print("\n-- Available Methods to call --")
        print("client.invoke_method(method_name, args, optional callback function)")
        print("-------------------------------------------------------------------")
        for method in self.state["methods"].values():
            if not "noo::" in method.info.name:
                print(method)


    def shutdown(self):
        """Method for shutting down Websocket connection"""
        
        asyncio.run_coroutine_threadsafe(self._socket.close(), self._loop)

    



"""Module for Handling Raw Messages from the Server"""

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from penne.delegates import Delegate
    from penne.messages import Message
    from penne.core import Client

import weakref

from . import messages

# Helper Methods
def handle_update(client, message: Message, specifier: str):
    """Update a delegate in the current state

    Args:
        client (Client): 
            client to be updated
        message (Message): 
            message containing updates
        specifier (str): 
            which part of state to update
    """

    current_state = client.state[specifier][message.id]
    for attribute, value in message.as_dict().items():
        setattr(current_state, attribute, value)


def delegate_from_context(client: Client, context: Message) -> Delegate:
    """Get delegate object from a context message object
    
    Args:
        client (Client): client to get delegate from
        context (Message): object containing context
    
    Raises:
        Exception: Couldn't get delegate from context
    """

    if not context:
        target_delegate = client.state["document"]
    elif hasattr(context, "table"):
        target_delegate = client.state["tables"][context.table]
    elif hasattr(context, "entity"):
        target_delegate = client.state["entities"][context.entity]
    elif hasattr(context, "plot"):
        target_delegate = client.state["plots"][context.plot]
    else:
        raise Exception("Couldn't get delegate from context")
    
    return target_delegate


def handle(client: Client, id, message_dict):
    """Handle message from server

    'Handle' uses the ID attached to message to get handling info, and uses this info 
    to take proper course of action with message. The function has 5 main sections 
    handling create, delete, and update messages along with signalinvocation and reply
    messages. For now all other communication messages are simply printed.

    'Handle' is also responsible for managing the client's state and working with the
    delegates in a couple of key ways. This function creates, deletes, and updates
    delegates as well as invoking methods on the delegates using signals.

    Args:
        client (Client): client receiving the message
        encoded_message (CBOR array): array with id and message as dictionary
    """
    
    # Process message using ID from dict
    handle_info = client.server_message_map[id]
    action = handle_info.action
    specifier = handle_info.specifier
    message_obj: Message = messages.Message.from_dict(message_dict)
    print(f"Message: {action} {specifier}")

    if specifier == "plots":
        print(f"\n  {action} - {specifier}\n{message_obj}")
    
    # Update state based on map info
    if action == "create":

        # Create instance of delegate
        specifier = specifier
        reference = weakref.ref(client)
        reference_obj = reference()
        delegate: Delegate = client.delegates[specifier](reference_obj, message_obj, specifier)

        # Update state and pass message info to the delegate's handler
        client.state[specifier][message_obj.id] = delegate
        delegate.on_new(message_obj)
    
    elif action == "delete":

        state_delegate: Delegate = client.state[specifier][message_obj.id]

        # Update delegate and state
        state_delegate.on_remove(message_obj)
        del state_delegate

    elif action == "update":

        if specifier != "document":
            handle_update(client, message_obj, specifier)
            client.state[specifier][message_obj.id].on_update(message_obj)
        else:
            client.state[specifier].on_update(message_obj)

    elif action == "reply":

        # Handle callback functions
        if hasattr(message_obj, "method_exception"):
            raise Exception(f"Method call ({message_obj.invoke_id}) resulted in exception from server: {message_obj.method_exception}")
        else:
            callback = client.callback_map.pop(message_obj.invoke_id)
            if callback:
                callback(message_obj.result) if hasattr(message_obj, "result") else callback()


    elif action == "invoke":

        # Handle invoke message from server
        signal_data = message_obj.signal_data
        signal: Delegate = client.state["signals"][message_obj.id]

        # Determine the delegate the signal is being invoked on
        context = getattr(message_obj, "context", False)
        target_delegate = delegate_from_context(client, context)

        # Invoke signal attached to target delegate
        print(f"Invoking {signal.info.name} w/ args: {signal_data}")
        target_delegate.signals[signal.info.name](*signal_data)

    elif action == "initialized":

        if client.on_connected:
            client.on_connected()

    else:
        # Document reset messages
        print(message_obj)



"""Module Defining Default Delegates and Delegate Related Classes"""

from __future__ import annotations
from typing import TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from penne.messages import Message
    from penne.core import Client


class Delegate(object):
    """Parent class for all delegates
    
    Defines general methods that should be available for all delegates
    
    Attributes:
        _client (Client): Client delegate is attached to
        info (Message): Message containing all info on delegate
        specifier (str): Keyword specifying delegate in state
        signals (dict): Signals that can be called on delegate
        __all__ (list): Specify public methods, used in show_methods()
    """

    def __init__(self, client: Client, message: Message, specifier: str):
        self._client = client
        self.info = message
        self.specifier = specifier
        self.signals = {}
        self.__all__ = []
    
    def __repr__(self) -> str:
        return f"{self.specifier} delegate | {self.info.id}"

    # For all except Document Delegate
    def on_new(self, message: Message):
        pass

    # For Document, Table, Entity, Plot, Material, Light Delegates
    def on_update(self, message: Message):
        pass

    # For all except Document Delegate
    def on_remove(self, message: Message):
        pass


    def show_methods(self):
        """Show methods available on this delegate"""

        print(f"-- Methods on {self} --")
        print("--------------------------------------")
        for name in self.__all__:
            method = getattr(self, name)
            print(f">> '{name}'\n{method.__doc__}")


class InjectedMethod(object):
    """Class for representing injected method in delegate

    Attributes:
        method (method): method to be called
        injected (bool): attribute marking method as injected
    """

    def __init__(self, method_obj) -> None:
        self.method = method_obj
        self.injected = True

    def __call__(self, *args, **kwds):
        self.method(*args, **kwds)


class LinkedMethod(object):
    """Class linking target delegate and method's delegate 
        
    make a cleaner function call in injected method
    
    Attributes:
        _obj_delegate (delegate): 
            delgate method is being linked to
        _method_delegate (MethodDelegate): 
            the method's delegate 
    """

    def __init__(self, object_delegate: Delegate, method_delegate: Delegate):
        self._obj_delegate = object_delegate
        self._method_delegate = method_delegate

    def __call__(self, on_done=None, *arguments):
        self._method_delegate.invoke(self._obj_delegate, arguments, callback=on_done)


def inject_methods(delegate: Delegate, methods: list):
    """Inject methods into a delegate class

    Args:
        delegate_name (str): 
            identifier for delegate to be modified
        methods (list): 
            list of method id's to inject
    """

    # Clear out old injected methods
    for name in dir(delegate):
        att = getattr(delegate, name)
        if hasattr(att, "injected"):
            print(f"Deleting: {name} in inject methods")
            delattr(delegate, name)

    state_methods = delegate._client.state["methods"] 
    for id in methods:

        # Get method delegate and manipulate name to exclude noo::
        method = state_methods[tuple(id)]
        name = method.info.name[5:]

        # Create injected by linking delegates, and creating call method
        linked = LinkedMethod(delegate, method)
        injected = InjectedMethod(linked.__call__)

        setattr(delegate, name, injected)


def inject_signals(delegate: Delegate, signals: list[list[int]]):
    """Method to inject signals into delegate

    Args:
        delegate (delegate): 
            delegate object to be injected 
        signals (list): 
            list of signal id's to be injected
    """

    state_signals = delegate._client.state["signals"]
    injected_signals = {}
    for id in signals:
        signal: SignalDelegate = state_signals[tuple(id)]
        injected_signals[signal.info.name] = None
    delegate.signals = injected_signals



"""
Default Delegates for Python Client
"""

class MethodDelegate(Delegate):
    """Delegate representing a method which can be invoked on the server

    Attributes:
        _client (client object): 
            client delegate is a part of 
        info (message): 
            message containing information on the method
        specifier (str): 
            keyword for specifying the type of delegate
        context_map (dict):
            mapping specifier to context for method invocation
    """

    def __init__(self, client: Client, message: Message, specifier: str):
        self._client = client
        self.info = message
        self.specifier = specifier
        self.context_map = {
            "tables": "table",
            "plots": "plot",
            "entities": "entity"
        }

    def on_new(self, message: Message):
        pass

    def on_remove(self, message: Message):
        pass

    def invoke(self, on_delegate: Delegate, args=None, callback=None):
        """Invoke this delegate's method

        Args:
            on_delegate (delegate):
                delegate method is being invoked on 
                used to get context
            args (list, optional):
                args for the method
            callback (function):
                function to be called when complete
        """
        context = {self.context_map[on_delegate.specifier]: on_delegate.info.id}
        self._client.invoke_method(self.info.id, args, context=context, on_done=callback)


    def __repr__(self) -> str:
        """Custom string representation for methods"""
        
        rep = f"{self.info.name}:\n\t{self.info.doc}\n\tReturns: {self.info.return_doc}\n\tArgs:"
        for arg in self.info.arg_doc:
            rep += f"\n\t\t{arg.name}: {arg.doc}"
        return rep


class SignalDelegate(Delegate):
    """Delegate representing a signal coming from the server

    Attributes:
        _client (Client): 
            client delegate is a part of 
        info (message): 
            message containing information on the signal
        specifier (str): 
            keyword for specifying the type of delegate
    """
    
    def __init__(self, client: Client, message: Message, specifier: str):
        self._client = client
        self.info = message
        self.specifier = specifier

    def on_new(self, message: Message):
        pass

    def on_remove(self, message: Message): 
        pass


class SelectionRange(tuple):
    """Selection of range of rows"""

    def __new__(cls, key_from: int, key_to: int):
        return super().__new__(SelectionRange, (key_from, key_to))


class Selection(object):
    """Selection of certain rows in a table

    Attributes:
        name (str): 
            name of the selection
        rows (list[int]): 
            list of indices of rows
        row_ranges (list[SelectionRange]): 
            ranges of selected rows
    """

    def __init__(self, name: str, rows: list[int] = None, row_ranges: list[SelectionRange] = None):
        self.name = name
        self.rows = rows
        self.row_ranges = row_ranges

    def __repr__(self) -> str:
        return f"Selection Object({self.__dict__})"

    def __getitem__(self, attribute):
        return getattr(self, attribute)


class TableDelegate(Delegate):
    """Delegate representing a table

    Each table delegate corresponds with a table on the server
    To use the table, you must first subscribe 

    Attributes:
        _client (Client): 
            weak ref to client to invoke methods and such
        dataframe (Dataframe): 
            dataframe representing current state of the table
        selections (dict): 
            mapping of name to selection object
        signals (signals): 
            mapping of signal name to function
        name (str): 
            name of the table
        id (list): 
            id group for delegate in state and table on server
    """

    def __init__(self, client: Client, message: Message, specifier: str):
        super().__init__(client, message, specifier)
        self.name = "Table Delegate"
        self.selections = {}
        self.signals = {
            "tbl_reset" : self._reset_table,
            "tbl_rows_removed" : self._remove_rows,
            "tbl_updated" : self._update_rows,
            "tbl_selection_updated" : self._update_selection
        }
        # Specify public methods 
        self.__all__ = [
            "subscribe", 
            "request_clear", 
            "request_insert", 
            "request_remove", 
            "request_update", 
            "request_update_selection",
            "plot"
        ]
        self.plotting = None


    def _on_table_init(self, init_info: Message, on_done=None):
        """Creates table from server response info

        Args:
            init_info (Message Obj): 
                Server response to subscribe which has columns, keys, data, 
                and possibly selections
        """

        # Extract data from init info and transpose rows to cols
        row_data = getattr(init_info, "data")
        cols = getattr(init_info, "columns")
        print(f"Table Initialized with cols: {cols} and row data: {row_data}")


    def _reset_table(self):
        """Reset dataframe and selections to blank objects

        Method is linked to 'tbl_reset' signal
        """

        self.selections = {}


    def _remove_rows(self, key_list: list[int]):
        """Removes rows from table

        Method is linked to 'tbl_rows_removed' signal

        Args:
            key_list (list): list of keys corresponding to rows to be removed
        """

        print(f"Removed Rows: {key_list}...\n", self.dataframe)



    def _update_rows(self, keys: list[int], rows: list):
        """Update rows in table

        Method is linked to 'tbl_updated' signal

        Args:
            keys (list): 
                list of keys to update
            cols (list): 
                list of cols containing the values for each new row,
                should be col for each col in table, and value for each key
        """

        print(f"Updated Rows...{keys}\n")
        

    def _update_selection(self, selection_obj: Selection):
        """Change selection in delegate's state to new selection object

        Method is linked to 'tbl_selection_updated' signal

        Args:
            selection_obj (Selection): 
                obj with new selections to replace obj with same name
        """

        self.selections[selection_obj.name] = selection_obj
        print(f"Made selection {selection_obj.name} = {selection_obj}")


    def _relink_signals(self):
        """Relink the signals for built in methods

        These should always be linked, along with whatever is injected,
        so relink on new and on update messages
        """

        self.signals["noo::tbl_reset"] = self._reset_table
        self.signals["noo::tbl_rows_removed"] = self._remove_rows
        self.signals["noo::tbl_updated"] = self._update_rows
        self.signals["noo::tbl_selection_updated"] = self._update_selection


    def on_new(self, message: Message):
        """Handler when create message is received

        Args:
            message (Message): create message with the table's info
        """
        
        # Set name
        name = message["name"]
        methods = message["methods_list"]
        signals = message["signals_list"]
        if name: self.name = name
    
        # Inject methods and signals
        if methods: inject_methods(self, methods)
        if signals: inject_signals(self, signals)

        # Reset
        self._reset_table()
        self._relink_signals()


    def on_update(self, message: Message):
        """Handler when update message is received
        
        Args:
            message (Message): update message with the new table's info
        """

        self._relink_signals()
    

    def on_remove(self, message: Message):
        pass


    def subscribe(self, on_done: Callable=None):
        """Subscribe to this delegate's table

        Calls on_table_init as callback
        
        Raises:
            Exception: Could not subscribe to table
        """

        try:
            # Allow for calback after table init
            lam = lambda data: self._on_table_init(data, on_done)
            self.tbl_subscribe(on_done=lam)
        except:
            raise Exception("Could not subscribe to table")

    
    def request_insert(self, row_list: list[list[int]], on_done=None):
        """Add rows to end of table

        User endpoint for interacting with table and invoking method
        For input, row list is list of rows. Also note that tables have
        nine columns by default (x, y, z, r, g, b, sx, sy, sz).
        x, y, z -> coordinates
        r, g, b -> color values [0, 1]
        sx, sy, sz -> scaling factors, default size is 1 meter

        Row_list: [[1, 2, 3, 4, 5, 6, 7, 8, 9]]

        Args:
            col_list (list, optional): add rows as list of columns
            row_list (list, optional): add rows using list of rows
            on_done (function, optional): callback function
        Raises:
            Invalid input for request insert exception
        """

        self.tbl_insert(on_done, row_list)
    

    def request_update(self, keys:list[int], rows:list[list[int]], on_done=None):
        """Update the table using a DataFrame

        User endpoint for interacting with table and invoking method

        Args:
            data_frame (DataFrame):
                data frame containing the values to be updated
            on_done (function, optional): 
                callback function called when complete
        """
        
        self.tbl_update(on_done, keys, rows)


    def request_remove(self, keys: list[int], on_done=None):
        """Remove rows from table by their keys

        User endpoint for interacting with table and invoking method

        Args:
            keys (list):
                list of keys for rows to be removed
            on_done (function, optional): 
                callback function called when complete
        """

        self.tbl_remove(on_done, keys)


    def request_clear(self, on_done=None):
        """Clear the table

        User endpoint for interacting with table and invoking method

        Args:
            on_done (function, optional): callback function called when complete
        """
        self.tbl_clear(on_done)


    def request_update_selection(self, name: str, keys: list[int], on_done=None):
        """Update a selection object in the table

        User endpoint for interacting with table and invoking method

        Args:
            name (str):
                name of the selection object to be updated
            keys (list):
                list of keys to be in new selection
            on_done (function, optional): 
                callback function called when complete
        """

        self.tbl_update_selection(on_done, name, {"rows": keys})


class DocumentDelegate(Delegate):
    pass
    
class EntityDelegate(Delegate):
    pass

class PlotDelegate(Delegate):
    pass

class MaterialDelegate(Delegate):
    pass

class GeometryDelegate(Delegate):
    pass

class LightDelegate(Delegate):
    pass

class ImageDelegate(Delegate):
    pass

class TextureDelegate(Delegate):
    pass

class SamplerDelegate(Delegate):
    pass

class BufferDelegate(Delegate):
    pass

class BufferViewDelegate(Delegate):
    pass



"""Client Test Script

Test the functionality of the client using a custom delegate and callback functions.
Designed to interact with PlottyN server to create a 3d scatter chart
"""

from typing import Callable
import unittest
import multiprocessing

import pandas as pd
import matplotlib.pyplot as plt

from penne.client import create_client
from penne.delegates import TableDelegate



def get_plot_data(df: pd.DataFrame):
    """Helper function to extract data for the plot from the dataframe"""

    data = {
        "xs": df["x"], 
        "ys": df["y"], 
        "zs": df["z"],
        "s" : [(((sx + sy + sz) / 3) * 1000) for sx, sy, sz in zip(df["sx"], df["sy"], df["sz"])],
        "c" : [(r, g, b) for r, g, b in zip(df["r"], df["g"], df["b"])]
    }
    return data


def on_close(event):
    """Event handler for when window is closed"""

    plt.close('all')

def plot_process(df: pd.DataFrame, receiver):
    """Process for plotting the table as a 3d scatter plot

    Args:
        df (DataFrame): 
            the data to be plotted
        receiver (Pipe connection object):
            connection to receive updates from root process
    """

    # Enable interactive mode
    plt.ion()

    # Make initial plot
    fig = plt.figure()
    fig.canvas.mpl_connect('close_event', on_close)
    ax = fig.add_subplot(projection='3d')
    data = get_plot_data(df)
    ax.scatter(**data)

    ax.set_xlabel('X Label')
    ax.set_ylabel('Y Label')
    ax.set_zlabel('Z Label')

    plt.draw()
    plt.pause(.001)

    # Update loop
    while True:

        # If update received, redraw the scatter plot
        if receiver.poll(.1):
            update = receiver.recv()
            plt.cla() # efficient? better way to set directly?
            ax.scatter(**update)
            ax.set_xlabel('X Label')
            ax.set_ylabel('Y Label')
            ax.set_zlabel('Z Label') 
            plt.pause(.001)

        # Keep GUI event loop going as long as window is still open
        elif plt.fignum_exists(fig.number):
            plt.pause(.01)
        else:
            break


class TestDelegate(TableDelegate):
    """Overide Table Delegate to Add Plotting Capabilities"""

    def __init__(self, client, message, specifier) -> None:
        super().__init__(client, message, specifier)
        self.dataframe: pd.DataFrame = None

    def _on_table_init(self, init_info, on_done=None):
        """Creates table from server response info

        Args:
            init_info (Message Obj): 
                Server response to subscribe which has columns, keys, data, 
                and possibly selections
        """

        # Extract data from init info and transpose rows to cols
        row_data = getattr(init_info, "data")
        col_data = [list(i) for i in zip(*row_data)]
        cols = getattr(init_info, "columns")
        data_dict = {getattr(col, "name"): data for col, data in zip(cols, col_data)}
        
        self.dataframe = pd.DataFrame(data_dict, index=getattr(init_info, "keys"))

        # Initialize selections if any
        selections = getattr(init_info, "selections", [])
        for selection in selections:
            self.selections[selection.name] = selection
        
        print(f"Initialized data table...\n{self.dataframe}")
        if on_done: on_done()


    def _reset_table(self):
        """Reset dataframe and selections to blank objects

        Method is linked to 'tbl_reset' signal
        """

        self.dataframe = pd.DataFrame()
        self.selections = {}

        if self.plotting:
            self._update_plot()


    def _remove_rows(self, key_list: list[int]):
        """Removes rows from table

        Method is linked to 'tbl_rows_removed' signal

        Args:
            key_list (list): list of keys corresponding to rows to be removed
        """

        self.dataframe.drop(index=key_list, inplace=True)
        print(f"Removed Rows: {key_list}...\n", self.dataframe)

        if self.plotting:
            self._update_plot()


    def _update_rows(self, keys: list[int], rows: list):
        """Update rows in table

        Method is linked to 'tbl_updated' signal

        Args:
            keys (list): 
                list of keys to update
            cols (list): 
                list of cols containing the values for each new row,
                should be col for each col in table, and value for each key
        """

        for key, row in zip(keys, rows):
            self.dataframe.loc[key] = row

        if self.plotting:
            self._update_plot()
    
        print(f"Updated Rows...{keys}\n", self.dataframe)


    def get_selection(self, name: str):
        """Get a selection object and construct Dataframe representation

        Args:
            name (str) : name of selection object to get
        """
        # Try to retrieve selection object from instance or return blank frame
        try:
            sel_obj = self.selections[name]
        except:
            return pd.DataFrame(columns=self.dataframe.columns)

        frames = []

        # Get rows already in that selection
        if sel_obj.rows:
            frames.append(self.dataframe.loc[sel_obj["rows"]])

        # Uses ranges in object to get other rows
        if sel_obj.row_ranges:
            ranges = sel_obj["row_ranges"]
            for r in ranges:
                frames.append(self.dataframe.loc[r[0]:r[1]-1])

        # Return frames concatenated
        df = pd.concat(frames)
        print(f"Got selection for {sel_obj}\n{df}")
        return df

    def _update_plot(self):
        """Update plotting process when dataframe is updated"""

        df = self.dataframe
        self.sender.send(get_plot_data(df))

    def plot(self, on_done: Callable=None):
        """Creates plot in a new window

        Uses matplotlib to plot a representation of the table
        """

        self.sender, receiver = multiprocessing.Pipe()

        self.plotting=multiprocessing.Process(target=plot_process, args=(self.dataframe, receiver))
        self.plotting.start()
        if on_done: on_done()


class Tests(unittest.TestCase):
    
    def test(self):

        # Create callback functions
        # Not sure about 'response' - cleaner way?
        args = [[1,2,3,4,5],[1,2,3,4,5],[1,2,3,4,5],[(0,1,1),(0,1,1),(0,1,1),(0,1,1),(0,1,1)]]
        create_table = lambda: client.invoke_method("new_point_plot", args, on_done=subscribe)
        subscribe = lambda response: client.state["tables"][(0, 0)].subscribe(on_done=plot)
        plot = lambda: client.state["tables"][(0, 0)].plot(on_done=insert_points)
        insert_points = lambda: client.state["tables"][(0, 0)].request_insert(
            row_list=[[8, 8, 8, .3, .2, 1, .05, .05, .05],[9,9,9,.1,.2,.5,.02,.02,.02, "Annotation"]], 
            on_done=update_rows
            )
        update_rows = lambda response: client.state["tables"][(0, 0)].request_update([3],[[4,6,3,0,1,0,.1,.1,.1,"Updated this row"]], on_done=get_selection)
        get_selection = lambda response: client.state["tables"][(0, 0)].request_update_selection("Test Select", [1, 2, 3], on_done=remove_row)
        remove_row = lambda response: client.state["tables"][(0, 0)].request_remove([2], on_done=shutdown)
        shutdown = lambda response: client.shutdown()

        # Creat client and start callback chain
        del_hash = {"tables" : TestDelegate}
        client = create_client("ws://localhost:50000", del_hash, on_connected=create_table)

        # Wait for client thread to finish
        client.thread.join()
        print(f"Finished Testing")

if __name__ == "__main__":
    unittest.main()